/*
** Boot frame buffer console routines
**
** Provides simple frame buffer support for boot banner, kernel command
** line editing, error messages, etc.
**
** Copyright 2002 Jochen Roth (jochen@panix.com)
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License,
** Version 2, as published by the Free Software Foundation.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <video/font.h>

#include "bfbcon.h"

#define	SOLID_CURSOR 1

#ifdef CONFIG_FB
#include "../../../../drivers/video/font_sun8x16.c"
/* extern struct fbcon_font_desc font_vga_8x16; */
#endif

#ifdef CONFIG_VGA
/* use vga font from iso-font.h -- need to make that non-static. */
extern unsigned char font[];
static struct fbcon_font_desc font_vga = { 0, "", 8, 16, font, 0 };
#endif

static unsigned char *bfbcon_buffer, *bfbcon_font_data;
static int bfbcon_depth, bfbcon_dark, bfbcon_lite, bfbcon_cursor_xor;
static int bfbcon_font_width, bfbcon_font_height, bfbcon_font_bpc;
static int bfbcon_width, bfbcon_height, bfbcon_rows, bfbcon_cols,
		bfbcon_row, bfbcon_col, bfbcon_bytes_per_row;

static void bfbcon_cursor(void);

void bfbcon_init(void *fbmem, int width, int height, int depth,
		int dark, int lite, struct fbcon_font_desc *font)
{
	bfbcon_buffer = fbmem;
	bfbcon_width = width;
	bfbcon_height = height;
	bfbcon_depth = depth / 8;
	bfbcon_dark = dark;
	bfbcon_lite = lite;
	bfbcon_cursor_xor = dark ^ lite;

#ifdef CONFIG_FB
	if(!font) font = &font_sun_8x16;
#endif
#ifdef CONFIG_VGA
	if(!font) font = &font_vga;
#endif
	if(font) {
		bfbcon_font_width = font->width;
		bfbcon_font_height = font->height;
		bfbcon_font_data = font->data;
	}

	bfbcon_font_bpc = ((bfbcon_font_width + 7) / 8) * bfbcon_font_height;
	bfbcon_bytes_per_row = bfbcon_font_height * bfbcon_width * bfbcon_depth;

	bfbcon_rows = bfbcon_height / bfbcon_font_height;
	bfbcon_cols = bfbcon_width / bfbcon_font_width;

	bfbcon_row = bfbcon_rows - 5;
	bfbcon_col = 0;
	bfbcon_cursor();
}


static void bfbcon_put_pixel(unsigned char *bp, int p)
{
	switch(bfbcon_depth) {
		case 1:
			*bp = (unsigned char)p;
			break;
		default:
			break;
	}
}

static void bfbcon_set_pixel(int x, int y, int p)
{
	unsigned char *bp = bfbcon_buffer + (y * bfbcon_width + x) * bfbcon_depth;
	bfbcon_put_pixel(bp, p);
}

static int bfbcon_get_pixel(int x, int y)
{
	unsigned char *bp = bfbcon_buffer + (y * bfbcon_width + x) * bfbcon_depth;

	switch(bfbcon_depth) {
		case 1:
			return(*bp);
		default:
			break;
	}
	return(0);
}

void	bfbcon_plot(int x, int y, int lite)
{
	if(bfbcon_buffer)
		bfbcon_set_pixel(x, y, lite ? bfbcon_lite : bfbcon_dark);
}

static void bfbcon_paintc(unsigned c)
{
	unsigned char *bp;
	unsigned char *fp = bfbcon_font_data + bfbcon_font_bpc * c;
	int	row, col;

	bp = bfbcon_buffer + bfbcon_row * bfbcon_bytes_per_row
		+ bfbcon_col * bfbcon_font_width * bfbcon_depth;

	for(row = 0; row < bfbcon_font_height; row++) {
		unsigned char *fpp = fp, *bpp = bp;
		unsigned char pix = 0;
		int bit;
		for(col = bit = 0; col < bfbcon_font_width; col++) {
			if(bit == 0) {
				pix = *fpp++;
				bit = 8;
			}
#if 0	/* jr */
			if(pix & (1 << --bit))
				*bpp++ = (unsigned char)bfbcon_lite;
			else
				*bpp++ = (unsigned char)bfbcon_dark;
#endif	/* jr */
			bfbcon_put_pixel(bpp++, (pix & (1 << --bit)) ? bfbcon_lite : bfbcon_dark);
		}
		fp += ((bfbcon_font_width + 7) / 8);
		bp += bfbcon_width * bfbcon_depth;
	}

}

static void bfbcon_cursor_flip(int x, int y)
{
	int pix;

	pix = bfbcon_get_pixel(x, y);
	pix ^= bfbcon_cursor_xor;
	bfbcon_set_pixel(x, y, pix);
}

static void bfbcon_cursor(void)
{
	int	x, y, index;

	x = bfbcon_col * bfbcon_font_width;
	y = bfbcon_row * bfbcon_font_height;

#if SOLID_CURSOR
	for(index = 0; index < bfbcon_font_height; index++) {
		int w;
		for(w = 0; w < bfbcon_font_width; w++) {
			bfbcon_cursor_flip(x+w, y+index);
		}
	}
#else
	for(index = 0; index < bfbcon_font_width; index++) {
		bfbcon_cursor_flip(x+index, y);
		bfbcon_cursor_flip(x+index, y+bfbcon_font_height-1);
	}
	for(index = 1; index < bfbcon_font_height-1; index++) {
		bfbcon_cursor_flip(x, y+index);
		bfbcon_cursor_flip(x+bfbcon_font_width-1, y+index);
	}
#endif
}

static void bfbcon_memmove(unsigned char *d, unsigned char *s, int count)
{
	while(count >= 4) {
		*((unsigned *)d) = *((unsigned *)s);
		d += 4;
		s += 4;
		count -= 4;
	}
	while(count--)
		*d++ = *s++;
}

static void bfbcon_memset(unsigned char *p, unsigned char v, int count)
{
	while(count--)
		*p++ = v;
}

static void bfbcon_scroll(void)
{
	bfbcon_memmove(bfbcon_buffer,
		bfbcon_buffer + bfbcon_width * bfbcon_depth * bfbcon_font_height,
		bfbcon_width * (bfbcon_height - bfbcon_font_height) * bfbcon_depth);
	bfbcon_memset(bfbcon_buffer + bfbcon_width * (bfbcon_height-bfbcon_font_height) * bfbcon_depth,
		bfbcon_dark, bfbcon_width * bfbcon_depth * bfbcon_font_height);
}

void	bfbcon_moveto(int row, int col)
{
	if(!bfbcon_buffer) return;
	bfbcon_cursor();
	bfbcon_row = row;
	if(bfbcon_row < 0)
		bfbcon_row = 0;
	if(bfbcon_row >= bfbcon_rows)
		bfbcon_row = bfbcon_rows - 1;
	bfbcon_col = col;
	if(bfbcon_col < 0)
		bfbcon_col = 0;
	if(bfbcon_col >= bfbcon_cols)
		bfbcon_col = bfbcon_cols - 1;
	bfbcon_cursor();
}

void	bfbcon_putc(unsigned char c)
{
	if(!bfbcon_buffer) return;

	bfbcon_cursor();

	if(c == '\n') {
		bfbcon_col = 0;
		bfbcon_row++;
	} else if(c == '\r') {
		bfbcon_col = 0;
	} else if(c == '\b') {
		if(bfbcon_col) {
			bfbcon_col--;
		}
	} else {
		bfbcon_paintc((unsigned)c);
		if(++bfbcon_col >= bfbcon_cols) {
			bfbcon_col = 0;
			bfbcon_row++;
		}
	}

	if(bfbcon_row >= bfbcon_rows) {
		bfbcon_row = bfbcon_rows - 1;
		bfbcon_scroll();
	}
	bfbcon_cursor();
}

void	bfbcon_puts(unsigned char *s)
{
	while(*s)
		bfbcon_putc(*s++);
}



