/*
 *
 *  85xx_dma_common.c, version 1.0
 *
 *  Based on hhnet_dma from MontaVista
 *
 *  This software may be used and distributed according to the terms of
 *  the GNU Public License, Version 2, incorporated herein by reference.
 *
 *  Contact:  <kumar.gala@motorola.com>
 *
 */

#include "85xx_dma_priv.h"
#include "85xx_dma.h"

#define DRIVER_VERSION "1.0"

MODULE_AUTHOR("Kumar Gala");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("85xx General Purpose DMA engine API");

unsigned int noisegen = 0;
MODULE_PARM(noisegen, "i");
MODULE_PARM_DESC(noisegen, "boolean param - if set, we create a circular "
		 "list of BDs that never stop transferring data - only"
		 "works if not striding");

static struct dma_channel channels[DMA_NUM_CHANNELS];
static void mpc85xx_dma_irq_handler(int irq, void *dev, struct pt_regs *regs);

/* parameters/command-line options */
static int pool_size = MAX_POOL_SIZE;
static int irq = DEFAULT_IRQ;

#if 0
/* kernel command line string handling */

static int __init
set_pool_size(char *str)
{

	pool_size = simple_strtoul(str, NULL, 0);
	return 1;
}

__setup("mpc85xx_dma_poolsize=", set_pool_size);

static int __init
set_dma_irq(char *str)
{

	irq = simple_strtoul(str, NULL, 0);
	return 1;
}

__setup("mpc85xx_dma_irq=", set_dma_irq);
#endif

/* caller must hold ch->lock */
static void
__mpc85xx_dma_start(struct dma_channel *ch, struct list_head *head)
{
	struct dma_reg *pdma = ch->pdma;
	struct dma_link_desc *last_link;
	struct dma_list_desc *last_list;
	Debug((PRINT_PREFIX "__mpc85xx_dma_start\n"));

	while (pdma->SR & DMA_SR_CB) {

		udelay(100);
		printk((PRINT_PREFIX "Waiting for DMA to stop\n"));
	}

	pdma->MR &= ~DMA_MR_CS;

	/* clear all errors */
	pdma->SR = DMA_SR_ALL;

	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
		last_list = list_entry(head->prev, struct dma_list_desc, list);
		/* Set end of list */
		pdma->CLSDAR = last_list->next_list | DMA_CLNDAR_GLEN;
		last_list->next_list = DMA_LIST_DESC_EOLSD;
	} else {
		last_link = list_entry(head->prev, struct dma_link_desc,
				       pub.list);
		/* set end of link */
		pdma->CLNDAR = last_link->next_link | DMA_CLNDAR_GLEN;
		if (!noisegen) {
			last_link->next_link = DMA_LINK_DESC_EOLND;
		}
	}

	/*start transfer */
	pdma->MR |= DMA_MR_CS;
}

/* caller must hold ch->lock */
static void
__mpc85xx_dma_start_queue(struct dma_channel *ch)
{
	Debug((PRINT_PREFIX "__mpc85xx_dma_start_queue\n"));

	if (list_empty(&ch->active) && !list_empty(&ch->pending)) {
		list_splice(&ch->pending, &ch->active);
		INIT_LIST_HEAD(&ch->pending);
		__mpc85xx_dma_start(ch, &ch->active);
	}
}

static void
mpc85xx_dma_irq_handler(int irq, void *dev, struct pt_regs *regs)
{

	struct dma_channel *ch = (struct dma_channel *) dev;
	struct dma_reg *pdma = ch->pdma;
	struct dma_link_desc *p_linkdesc;
	struct dma_list_desc *p_listdesc;
	u32 status;

	Debug((PRINT_PREFIX "mpc85xx_dma_irq_handler\n"));

	/* read status from hardware */
	status = pdma->SR;

	if (!(status & (DMA_ERROR | DMA_SR_EOLNI | DMA_SR_EOLSI)))
		return;		/* Not us */

	Debug((PRINT_PREFIX "Status = 0x%08x\n", status));

	if (status & DMA_ERROR) {
		printk(PRINT_PREFIX "Status = 0x%08x\n", status);
	}

	/* clear all errors */
	pdma->SR = DMA_SR_ALL;

	/*
	 * pull the active list, save the hardware status in the first list
	 * element and add the list to the completed chain.
	 */
	spin_lock(&ch->lock);

	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
		/* reset end-of-chain marker in last list entry */
		p_listdesc =
		    list_entry(ch->active.prev, struct dma_list_desc, list);
		p_listdesc->next_list &= ~DMA_LIST_DESC_EOLSD;

		/*
		 * set end-of-list marker in first entry (to indicate to bottom half
		 * that it holds a valid status) and save the hardware status.
		 */

		p_listdesc =
		    list_entry(ch->active.next, struct dma_list_desc, list);
		p_listdesc->next_list |= DMA_LIST_DESC_EOLSD;
		p_listdesc->status = status;
	} else {
		/* reset end-of-chain marker in last list entry */
		p_linkdesc =
		    list_entry(ch->active.prev, struct dma_link_desc, pub.list);
		p_linkdesc->next_link &= ~DMA_LINK_DESC_EOLND;

		/*
		 * set end-of-list marker in first entry (to indicate to bottom half
		 * that it holds a valid status) and save the hardware status.
		 */
		p_linkdesc =
		    list_entry(ch->active.next, struct dma_link_desc, pub.list);
		p_linkdesc->next_link |= DMA_LINK_DESC_EOLND;
		p_linkdesc->pub.status = status;
	}
	/*
	 * move chain from active list to end of complete list and
	 * re-initialize active list
	 */
	list_splice(&ch->active, ch->complete.prev);
	INIT_LIST_HEAD(&ch->active);

	/* try to re-start the engine */
	__mpc85xx_dma_start_queue(ch);

	spin_unlock(&ch->lock);

	/* schedule the bottom half */
	queue_task(&ch->tq, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
}

int
mpc85xx_dma_queue_list(struct list_head *head, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	struct dma_link_desc *prev;
	struct list_head *pos;
	unsigned int flags;

	Debug((PRINT_PREFIX "mpc85xx_dma_queue_list\n"));

	if ((ch == NULL) || (head == NULL))
		return -EINVAL;

	/* If we are in striding mode we need to have called 
	 * mpc85xx_dma_set_striding_info first */
	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING)
		if (ch->curr_list == NULL)
			return -EINVAL;

	if (list_empty(head))
		return 0;

	prev = list_entry(head->prev, struct dma_link_desc, pub.list);

	/* For every public desc -- build the hardware link desc */
	list_for_each(pos, head) {
		struct dma_link_desc *pdesc =
		    list_entry(pos, struct dma_link_desc,
			       pub.list);
		int retval = mpc85xx_dma_build_desc(pdesc, prev, ch);

		if (retval)
			return retval;
		prev = pdesc;
	}

	spin_lock_irqsave(&ch->lock, flags);
	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
		/* THIS CODE NEEDS TO BE THOUGHT THRU - galak (8.6.03) */
		struct dma_link_desc *prev;
		struct dma_link_desc *next;
		struct dma_list_desc *ldesc;
		LIST_HEAD(head2);

		mpc85xx_dma_append_list(head, &ch->curr_list->link);
		prev =
		    list_entry(ch->curr_list->link.prev, struct dma_link_desc,
			       pub.list);
		next =
		    list_entry(ch->curr_list->link.next, struct dma_link_desc,
			       pub.list);
		prev->next_link = DMA_LINK_DESC_EOLND;

		ch->curr_list->first_link = next->paddr | DMA_CLNDAR_GLEN;

		list_add(&ch->curr_list->list, &head2);
		mpc85xx_dma_append_list_desc(&head2, &ch->pending);

		// automatically create a new list desc with the same striding info
		// as the last one. (removing this would just mean after a queue_list
		// the users would have to call set_striding_info themselves
		if (mpc85xx_dma_alloc_list_desc(&ldesc, ch) == 0) {
			ldesc->src_stride = ch->curr_list->src_stride;
			ldesc->dst_stride = ch->curr_list->dst_stride;
			ch->curr_list = ldesc;
			ch->curr_list->next_list =
			    ch->curr_list->paddr | DMA_CLNDAR_GLEN;
		} else {
			ch->curr_list = NULL;
		}
	} else {
		mpc85xx_dma_append_list(head, &ch->pending);
	}
	__mpc85xx_dma_start_queue(ch);
	spin_unlock_irqrestore(&ch->lock, flags);

	return 0;
}

static void
mpc85xx_dma_free_pool(struct dma_channel *ch)
{

	Debug((PRINT_PREFIX "mpc85xx_dma_free_pool\n"));

	if (ch->vpool_link == NULL)
		return;

	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING)
		if (ch->vpool_list == NULL)
			return;

	/* Free the link desc */
	pci_free_consistent(ch->pdev, ALIGNED_LINK_DESC_SIZE * pool_size,
			    ch->vpool_link, ch->ppool_link);
	ch->vpool_link = NULL;
	ch->ppool_link = (dma_addr_t) NULL;

	INIT_LIST_HEAD(&ch->free_link);

	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
		/* Free the list desc */
		pci_free_consistent(ch->pdev,
				    ALIGNED_LIST_DESC_SIZE * pool_size,
				    ch->vpool_list, ch->ppool_list);
		ch->vpool_list = NULL;
		ch->ppool_list = (dma_addr_t) NULL;
		ch->curr_list = NULL;

		INIT_LIST_HEAD(&ch->free_list);
	}
}

static int
mpc85xx_dma_init_pool(struct dma_channel *ch)
{
	struct dma_link_desc *p_linkdesc;
	struct dma_list_desc *p_listdesc;
	dma_addr_t paddr;
	int i;

	Debug((PRINT_PREFIX "mpc85xx_dma_init_pool\n"));

	if (ch->vpool_link != NULL)
		return 0;

	if (pool_size < MIN_POOL_SIZE) {
		printk(PRINT_PREFIX "pool_size = %d, MIN = %d, using MIN\n",
		       pool_size, MIN_POOL_SIZE);
		pool_size = MIN_POOL_SIZE;
	} else if (pool_size > MAX_POOL_SIZE) {
		printk(PRINT_PREFIX "pool_size = %d, MAX = %d, using MAX\n",
		       pool_size, MAX_POOL_SIZE);
		pool_size = MAX_POOL_SIZE;
	}

	ch->vpool_link =
	    pci_alloc_consistent(NULL, ALIGNED_LINK_DESC_SIZE * pool_size,
				 &ch->ppool_link);
	if (ch->vpool_link == NULL)
		return -ENOMEM;

	Debug((PRINT_PREFIX "vpool_link = %x, ppool_link = %x\n",
	       (unsigned int) ch->vpool_link, ch->ppool_link));

	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
		ch->vpool_list =
		    pci_alloc_consistent(NULL,
					 ALIGNED_LIST_DESC_SIZE * pool_size,
					 &ch->ppool_list);
		if (ch->vpool_list == NULL) {
			pci_free_consistent(ch->pdev,
					    ALIGNED_LINK_DESC_SIZE * pool_size,
					    ch->vpool_link, ch->ppool_link);
			return -ENOMEM;
		}

		Debug((PRINT_PREFIX "vpool_list = %x, ppool_list = %x\n",
		       (unsigned int) ch->vpool_list, ch->ppool_list));
	}

	p_linkdesc = (struct dma_link_desc *) ch->vpool_link;
	paddr = ch->ppool_link;

	for (i = 0; i < pool_size; i++) {

		p_linkdesc->paddr = paddr;

		list_add_tail(&p_linkdesc->pub.list, &ch->free_link);

		p_linkdesc =
		    (struct dma_link_desc *) ((u8 *) p_linkdesc +
					      ALIGNED_LINK_DESC_SIZE);
		paddr = (dma_addr_t) ((u8 *) paddr + ALIGNED_LINK_DESC_SIZE);
	}

	if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
		p_listdesc = (struct dma_list_desc *) ch->vpool_list;
		paddr = ch->ppool_list;

		for (i = 0; i < pool_size; i++) {
			INIT_LIST_HEAD(&p_listdesc->link);

			p_listdesc->paddr = paddr;

			list_add_tail(&p_listdesc->list, &ch->free_list);

			p_listdesc =
			    (struct dma_list_desc *) ((u8 *) p_listdesc +
						      ALIGNED_LIST_DESC_SIZE);
			paddr =
			    (dma_addr_t) ((u8 *) paddr +
					  ALIGNED_LIST_DESC_SIZE);
		}
	}

	return 0;
}

int
mpc85xx_dma_open(int num, int mode, void **dma_ch)
{

	int result;
	struct dma_channel *ch;

	Debug((PRINT_PREFIX "mpc85xx_dma_open\n"));

	if ((num < 0) || (num >= DMA_NUM_CHANNELS) || (dma_ch == NULL))
		return -EINVAL;

	ch = &channels[num];

	if (!ch->users) {
		ch->mode = mode;

		if ((result = mpc85xx_dma_init_pool(ch)) != 0)
			goto errout;

		/* stop channel, clear errors and interrupt indications */
		ch->pdma->MR = 0;
		ch->pdma->SR = DMA_SR_ALL;

		/* hook irq */
		if (request_irq(ch->irq, mpc85xx_dma_irq_handler, SA_SHIRQ,
				"mpc85xx_dma", (void *) ch) != 0) {
			mpc85xx_dma_free_pool(ch);
			printk(PRINT_PREFIX "Cannot get IRQ\n");
			result = -ENODEV;
			goto errout;
		}

		if ((ch->mode & DMA_MODE_STRIDING) == DMA_MODE_STRIDING) {
			/* enable programming error and end-of-link interrupts */
			ch->pdma->MR = DMA_MR_PEIE | DMA_MR_EOLSIE | DMA_MR_XFE;

			ch->tq.routine = mpc85xx_dma_striding_bh;
		} else {
			/* enable programming error and end-of-link interrupts */
			ch->pdma->MR = DMA_MR_PEIE | DMA_MR_EOLNIE;
			ch->tq.routine = mpc85xx_dma_bh;
		}
	} else {
		if ((mode & DMA_MODE_EXCLUSIVE) == DMA_MODE_EXCLUSIVE) {
			result = -EBUSY;
			goto errout;
		}
		if (mode != ch->mode) {
			result = -EBUSY;
			goto errout;
		}
	}
	ch->users++;

	*dma_ch = (void *) ch;
	return 0;

      errout:
	return result;
}

void
mpc85xx_dma_close(void *dma_ch)
{

	struct dma_channel *ch = (struct dma_channel *) dma_ch;

	Debug((PRINT_PREFIX "mpc85xx_dma_close\n"));

	if (ch == NULL)
		return;

	if (ch->users == 0)
		return;

	if (--ch->users == 0) {

		/* stop channel, clear errors, disable all interrupts */
		ch->pdma->MR = 0;
		ch->pdma->SR = DMA_SR_ALL;
		free_irq(ch->irq, ch);
		mpc85xx_dma_free_pool(ch);
	}
}

static void __devexit
mpc85xx_dma_remove_one(struct pci_dev *pdev)
{

	int i;

	Debug((PRINT_PREFIX "mpc85xx_dma_remove_one\n"));

	for (i = 0; i < DMA_NUM_CHANNELS; i++) {
		struct dma_channel *ch = &channels[i];

		if (ch->pdma != NULL) {
			iounmap((void *) ch->pdma);
			ch->pdma = NULL;
		}
	}
}

int __init
mpc85xx_dma_init(struct pci_dev *pdev)
{
	int i;

	Debug((PRINT_PREFIX "mpc85xx_dma_init_one\n"));
	Debug((PRINT_PREFIX "pool_size = %d\n", pool_size));
	Debug((PRINT_PREFIX "desc size = %d\n", sizeof (struct dma_link_desc)));
	Debug((PRINT_PREFIX "aligned desc size = %d\n",
	       ALIGNED_LINK_DESC_SIZE));

	for (i = 0; i < DMA_NUM_CHANNELS; i++) {
		struct dma_channel *ch = &channels[i];

		ch->pdev = 0;

		ch->irq = irq + i;

		ch->pdma = (struct dma_reg *) ioremap_nocache(get_ccsrbar() +
							      MPC85xx_DMA_OFFSET
							      + 0x80 * (i + 2),
							      0x80);
		if (ch->pdma == NULL) {
			printk(PRINT_PREFIX "Unable to map DMA registers\n");
			mpc85xx_dma_remove_one(pdev);
			return -EIO;
		}

		Debug((PRINT_PREFIX "v_pdma = 0x%08x\n",
		       (unsigned int) ch->pdma));

		ch->mode = DMA_MODE_SHARED;
		ch->users = 0;

		INIT_LIST_HEAD(&ch->free_link);
		INIT_LIST_HEAD(&ch->free_list);
		INIT_LIST_HEAD(&ch->pending);
		INIT_LIST_HEAD(&ch->active);
		INIT_LIST_HEAD(&ch->complete);

		ch->curr_list = NULL;

		spin_lock_init(ch->lock);
		ch->tq.routine = mpc85xx_dma_bh;
		ch->tq.data = (void *) ch;
		ch->vpool_link = NULL;
		ch->ppool_link = (dma_addr_t) NULL;
		ch->vpool_list = NULL;
		ch->ppool_list = (dma_addr_t) NULL;
	}
	return 0;
}

#if 0
void
debug_channel_info()
{
	int i = 0;
	for (i = 0; i < DMA_NUM_CHANNELS; i++) {
		printk("CHANNEL: %d\n", i);
		printk("   MODE: %d\n", channels[i].mode);
		printk("   USER: %d\n", channels[i].users);
		printk("\n");
	}
}

void
debug_link_info(char *s)
{
	int i = 0;
	struct dma_link_desc *p_linkdesc;
	if (channels[0].vpool_link == 0)
		return;
	printk("VPOOL = %p\n", p_linkdesc);
	p_linkdesc = (struct dma_link_desc *) channels[0].vpool_link;
	for (i = 0; i < 40; i++) {
		printk("%s: %p %p %08x %p %p\n", s, &p_linkdesc->paddr,
		       p_linkdesc->paddr, p_linkdesc->next_link,
		       p_linkdesc->pub.list.prev, p_linkdesc->pub.list.next);
		p_linkdesc =
		    (struct dma_link_desc *) ((u8 *) p_linkdesc +
					      ALIGNED_LINK_DESC_SIZE);
	}
}
#endif

static int
mpc85xx_dma_probe(struct ocp_device *ocp)
{

	Debug((PRINT_PREFIX "mpc85xx_dma_init_module\n"));

	return mpc85xx_dma_init(0);
}

static void
mpc85xx_dma_remove(struct ocp_device *ocp)
{
	Debug((PRINT_PREFIX "mpc85xx_cleanup_dma_module\n"));
}

static struct ocp_device_id moto_dma_ids[] = {
	{.vendor = OCP_VENDOR_MOTOROLA,.function = OCP_FUNC_DMA},
	{.vendor = OCP_VENDOR_INVALID}
};

static struct ocp_driver moto_dma_driver = {
	.name = "dma",
	.id_table = moto_dma_ids,

	.probe = mpc85xx_dma_probe,
	.remove = mpc85xx_dma_remove,
};

static int
    __init
mpc85xx_dma_init_module(void)
{
	printk(KERN_INFO "MOTO DMA driver v" DRIVER_VERSION "\n");
	if (ocp_register_driver(&moto_dma_driver) == 0) {
		/* No devices were bound to this driver */
		ocp_unregister_driver(&moto_dma_driver);
		return -ENODEV;
	}
	return 0;
}

static void
    __exit
mpc85xx_dma_cleanup_module(void)
{
	ocp_unregister_driver(&moto_dma_driver);
}

module_init(mpc85xx_dma_init_module);
module_exit(mpc85xx_dma_cleanup_module);

EXPORT_SYMBOL(mpc85xx_dma_queue_list);
EXPORT_SYMBOL(mpc85xx_dma_open);
EXPORT_SYMBOL(mpc85xx_dma_close);
EXPORT_SYMBOL(mpc85xx_dma_init);
