/*
 * arch/ppc/platforms/chestnut.c
 *
 * Board setup routines for IBM Chestnut
 *
 * Author: <source@mvista.com>
 *
 * <2004> (c) MontaVista Software, Inc. This file is licensed under
 * the terms of the GNU General Public License version 2. This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */

#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/seq_file.h>
#include <linux/ide.h>
#include <linux/serial.h>
#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <linux/irq.h>
#include <asm/hardirq.h>
#include <asm/machdep.h>
#include <asm/bootinfo.h>
#include <asm/mv64360.h>
#include <platforms/chestnut.h>
#include <platforms/chestnut_serial.h>

static u32 boot_base; /* Virtual addr of 8bit boot */
static u32 cpld_base; /* Virtual addr of CPLD Regs */

extern void gen550_progress(char *, unsigned short);
extern void gen550_init(int, struct serial_struct *);

/**************************************************************************
 * FUNCTION: chestnut_calibrate_decr
 *
 * DESCRIPTION: initialize decrementer interrupt frequency (used as system
 *              timer)
 *
 ****/
static void __init
chestnut_calibrate_decr(void)
{
	ulong freq;

	freq = CHESTNUT_BUS_SPEED / 4;

	printk("time_init: decrementer frequency = %lu.%.6lu MHz\n",
		freq/1000000, freq%1000000);

	tb_ticks_per_jiffy = freq / HZ;
	tb_to_us = mulhwu_scale_factor(freq, 1000000);
#ifdef CONFIG_ILATENCY
        {
                extern unsigned ticks_per_usec;
                ticks_per_usec = freq / 1000000;
        }
#endif

	return;
}

static int
chestnut_show_cpuinfo(struct seq_file *m)
{
	seq_printf(m, "vendor\t\t: IBM\n");
	seq_printf(m, "machine\t\t: 750FX/GX Eval Board (Chestnut/Buckeye)\n");

	return 0;
}

/**************************************************************************
 * FUNCTION: chestnut_find_end_of_memory
 * 
 * DESCRIPTION: ppc_md memory size callback
 *
 ****/
unsigned long __init
chestnut_find_end_of_memory(void)
{  
   	static int  mem_size = 0;

   	if (mem_size == 0) {
      		/* Next 2 lines are a kludge for mv64360_get_mem_size() */
      		mv64360_base = CHESTNUT_BRIDGE_REG_BASE;
      		mem_size = mv64360_get_mem_size();
   	}
   	return(mem_size);
}

#if defined(CONFIG_SERIAL)
static void __init
chestnut_early_serial_map(void)
{
	struct serial_struct serial_req;

	/* Setup serial port access */
	memset(&serial_req, 0, sizeof(serial_req));
	serial_req.baud_base = BASE_BAUD;
	serial_req.line = 0;
	serial_req.port = 0;
	serial_req.irq = UART0_INT;
	serial_req.flags = STD_COM_FLAGS;
	serial_req.io_type = SERIAL_IO_MEM;
	serial_req.iomem_base = ioremap(CHESTNUT_UART0_IO_BASE, 8);
	serial_req.iomem_reg_shift = 0;

#if defined(CONFIG_SERIAL_TEXT_DEBUG) || defined(CONFIG_KGDB)
	gen550_init(0, &serial_req);
#endif

	if (early_serial_setup(&serial_req) != 0)
		printk("Early serial init of port 0 failed\n");

	/* Assume early_serial_setup() doesn't modify serial_req */
	serial_req.line = 1;
	serial_req.port = 1;
	serial_req.irq = UART1_INT;
	serial_req.iomem_base = ioremap(CHESTNUT_UART1_IO_BASE, 8);

#if defined(CONFIG_SERIAL_TEXT_DEBUG) || defined(CONFIG_KGDB)
	gen550_init(1, &serial_req);
#endif

	if (early_serial_setup(&serial_req) != 0)
		printk("Early serial init of port 1 failed\n");
}
#endif

/**************************************************************************
 * FUNCTION: chestnut_map_irq
 *
 * DESCRIPTION: 0 return since PCI IRQs not needed
 * 
 ****/
static int __init
chestnut_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{  
	static char pci_irq_table[][4] = {
		{CHESTNUT_PCI_SLOT0_IRQ, CHESTNUT_PCI_SLOT0_IRQ,
		 CHESTNUT_PCI_SLOT0_IRQ, CHESTNUT_PCI_SLOT0_IRQ},
		{CHESTNUT_PCI_SLOT1_IRQ, CHESTNUT_PCI_SLOT1_IRQ,
		 CHESTNUT_PCI_SLOT1_IRQ, CHESTNUT_PCI_SLOT1_IRQ},
		{CHESTNUT_PCI_SLOT2_IRQ, CHESTNUT_PCI_SLOT2_IRQ,
		 CHESTNUT_PCI_SLOT2_IRQ, CHESTNUT_PCI_SLOT2_IRQ},
		{CHESTNUT_PCI_SLOT3_IRQ, CHESTNUT_PCI_SLOT3_IRQ,
		 CHESTNUT_PCI_SLOT3_IRQ, CHESTNUT_PCI_SLOT3_IRQ},
	};
	const long min_idsel = 1, max_idsel = 4, irqs_per_slot = 4;

	return (PCI_IRQ_TABLE_LOOKUP);
}


/**************************************************************************
 * FUNCTION: chestnut_setup_bridge
 *
 * DESCRIPTION: initalize board-specific settings on the MV64360
 *
 ****/
static mv64360_bridge_info_t info;

static void __init
chestnut_setup_bridge(void)
{
	int window;
	struct pci_controller *hose;
	u32 base, size;

   	if ( ppc_md.progress ) 
		ppc_md.progress("chestnut_setup_bridge: enter", 0);

   	memset(&info, 0, sizeof(mv64360_bridge_info_t));
   	MV64360_BRIDGE_INFO_DEFAULT(&info, chestnut_find_end_of_memory());

	/* disable hose B (PCI 1) */
	info.pci_1_mem_size = 0;
	info.pci_1_io_size = 0;

   	/* Lookup host bridge - on CPU 0 - no SMP support */
   	if (mv64360_find_bridges(CHESTNUT_BRIDGE_REG_BASE, &info,
               			chestnut_map_irq, 1)) {
        	printk("\n\nPCI Bridge initialization failed!\n");
   	}
    
   	/* Set up window for internal sram (256KByte insize) */
   	mv64360_cpu_internal_sram_set_window(CHESTNUT_INTERNAL_SRAM_BASE);

   	/*
    	 * Set up windows for boot CS, 32-bit flash, CPLD, UART and FRAM
	 * NOTE: chose the default for BootCS & CS0 location/order per the IBM
	 * docs (see System Controller Initialization U17 & U24)
    	 */

   	mv64360_cpu_boot_set_window(CHESTNUT_BOOT_8BIT_BASE, 
					CHESTNUT_BOOT_8BIT_SIZE);
	boot_base = (u32)ioremap(CHESTNUT_BOOT_8BIT_BASE, CHESTNUT_BOOT_8BIT_SIZE);

   	mv64360_cpu_dev_cs_set_window(0, CHESTNUT_32BIT_BASE, CHESTNUT_32BIT_SIZE);

   	mv64360_cpu_dev_cs_set_window(1, CHESTNUT_CPLD_BASE, 
					MV64360_WINDOW_SIZE_MIN);
	cpld_base = (u32)ioremap(CHESTNUT_CPLD_BASE, CHESTNUT_CPLD_SIZE);

   	mv64360_cpu_dev_cs_set_window(2, CHESTNUT_UART_BASE, 
					MV64360_WINDOW_SIZE_MIN);
   
   	mv64360_cpu_dev_cs_set_window(3, CHESTNUT_FRAM_BASE, 
					MV64360_WINDOW_SIZE_MIN);

	hose = info.hose_a;
	for (window = 0; window < MV64360_CPU_CS_DECODE_WINDOWS;
			window++) {
		mv64360_get_cpu_cs_base_and_size(window, &base,
				&size);

		mv64360_pci_slave_cs_set_window(hose, window,
				base, 0, base, 0, size);
	}

	mv64360_pci_slave_boot_set_window(hose,
		CHESTNUT_BOOT_SRAM_BASE, 0,
		CHESTNUT_BOOT_SRAM_BASE, CHESTNUT_BOOT_SRAM_SIZE);

	mv64360_pci_slave_dev_cs_set_window(hose, 0,
		CHESTNUT_32BIT_BASE, 0, CHESTNUT_32BIT_BASE, CHESTNUT_32BIT_SIZE);

	mv64360_pci_slave_dev_cs_set_window(hose, 1,
		CHESTNUT_32BIT_BASE, 0, CHESTNUT_32BIT_BASE, CHESTNUT_32BIT_SIZE);

	mv64360_pci_slave_dev_cs_set_window(hose, 2,
		CHESTNUT_BOOT_8BIT_BASE, 0,
		CHESTNUT_BOOT_8BIT_BASE, CHESTNUT_BOOT_8BIT_SIZE);

	mv64360_pci_slave_dev_cs_set_window(hose, 3,
		CHESTNUT_CPLD_BASE, 0,
		CHESTNUT_CPLD_BASE, CHESTNUT_CPLD_SIZE);

	mv64360_pci_slave_internal_sram_set_window(hose,
		CHESTNUT_INTERNAL_SRAM_BASE, 0,
		CHESTNUT_INTERNAL_SRAM_BASE, CHESTNUT_INTERNAL_SRAM_SIZE);

	mv64360_pci_slave_internal_registers_set_window(hose,
		CHESTNUT_BRIDGE_REG_BASE, 0,
		CHESTNUT_BRIDGE_REG_BASE, CHESTNUT_BRIDGE_REG_SIZE);

   	/*
    	 * Configure internal SRAM -
    	 * Cache coherent write back, incase 
	 *      CONFIG_MV64360_SRAM_CACHE_COHERENT set
    	 * Parity enabled.
    	 * Parity error propagation
    	 * Arbitration not parked for CPU only
    	 * Other bits are reserved.
    	 */
#ifdef CONFIG_MV64360_SRAM_CACHE_COHERENT
   	MV_REG_WRITE(MV64360_SRAM_CONFIG, 0x001600b2);
#else
   	MV_REG_WRITE(MV64360_SRAM_CONFIG, 0x001600b0);
#endif

   	/*
    	 * Setting the SRAM to 0. Note that this generates parity errors on 
	 * internal data path in SRAM since it's first time accessing it 
	 * while after reset it's not configured
    	*/
   	memset((void *)CHESTNUT_INTERNAL_SRAM_BASE, 0, CHESTNUT_INTERNAL_SRAM_SIZE);

	/*
	 * Configure MPP pins for PCI DMA
	 *
	 * PCI Slot	GNT pin		REQ pin
	 *	0	MPP16		MPP17
	 *	1	MPP18		MPP19
	 *	2	MPP20		MPP21
	 *	3	MPP22		MPP23
	 */
	MV_REG_WRITE(MV64360_MPP_CONTROL2,
			(0x1 << 0)  |	/* MPPSel16 PCI0_GNT[0] */
			(0x1 << 4)  |	/* MPPSel17 PCI0_REQ[0] */
			(0x1 << 8)  |	/* MPPSel18 PCI0_GNT[1] */
			(0x1 << 12) |	/* MPPSel19 PCI0_REQ[1] */
			(0x1 << 16) |	/* MPPSel20 PCI0_GNT[2] */
			(0x1 << 20) |	/* MPPSel21 PCI0_REQ[2] */
			(0x1 << 24) |	/* MPPSel22 PCI0_GNT[3] */
			(0x1 << 28));	/* MPPSel23 PCI0_REQ[3] */

	/*
	 * Set unused MPP pins for output, as per schematic note
	 *
	 * Unused Pins: MPP01, MPP02, MPP04, MPP05, MPP06
	 *		MPP09, MPP10, MPP13, MPP14, MPP15
	 */
	MV_RESET_REG_BITS(MV64360_MPP_CONTROL0,
			(0xf << 4)  |	/* MPPSel01 GPIO[1] */
			(0xf << 8)  |	/* MPPSel02 GPIO[2] */
			(0xf << 16) |	/* MPPSel04 GPIO[4] */
			(0xf << 20) |	/* MPPSel05 GPIO[5] */
			(0xf << 24));	/* MPPSel06 GPIO[6] */
	MV_RESET_REG_BITS(MV64360_MPP_CONTROL1,
			(0xf << 4)  |	/* MPPSel09 GPIO[9] */
			(0xf << 8)  |	/* MPPSel10 GPIO[10] */
			(0xf << 20) |	/* MPPSel13 GPIO[13] */
			(0xf << 24) |	/* MPPSel14 GPIO[14] */
			(0xf << 28));	/* MPPSel15 GPIO[15] */
	MV_SET_REG_BITS(MV64360_GPP_IO_CONTROL,
			BIT1  | BIT2  | BIT4  | BIT5  | BIT6  |
			BIT9  | BIT10 | BIT13 | BIT14 | BIT15);	/* Output */

   	/*
    	 * Configure the following MPP pins to indicate a level
    	 * triggered interrupt
    	 *
       	 * MPP24 - Board Reset (just map the MPP & GPP for chestnut_reset)
       	 * MPP25 - UART A  (high)
       	 * MPP26 - UART B  (high)
	 * MPP28 - PCI Slot 3 (low)
	 * MPP29 - PCI Slot 2 (low)
	 * MPP30 - PCI Slot 1 (low)
	 * MPP31 - PCI Slot 0 (low)
    	 */
        MV_RESET_REG_BITS(MV64360_MPP_CONTROL3,
                        BIT3 | BIT2 | BIT1 | BIT0 |	/* MPP 24 */
                        BIT7 | BIT6 | BIT5 | BIT4 |	/* MPP 25 */
                        BIT11 | BIT10 | BIT9 | BIT8 |	/* MPP 26 */
			BIT19 | BIT18 | BIT17 | BIT16 |	/* MPP 28 */
			BIT23 | BIT22 | BIT21 | BIT20 |	/* MPP 29 */
			BIT27 | BIT26 | BIT25 | BIT24 |	/* MPP 30 */
			BIT31 | BIT30 | BIT29 | BIT28);	/* MPP 31 */

   	/* 
	 * Define GPP 25 (high), 26 (high), 28 (low), 29 (low), 30 (low),
	 * 31 (low) interrupt polarity input signal and level triggered
    	 */
   	MV_RESET_REG_BITS(MV64360_GPP_LEVEL_CONTROL, BIT25 | BIT26);
   	MV_SET_REG_BITS(MV64360_GPP_LEVEL_CONTROL,
			BIT28 | BIT29 | BIT30 | BIT31);
   	MV_RESET_REG_BITS(MV64360_GPP_IO_CONTROL,
			BIT25 | BIT26 | BIT28 | BIT29 | BIT30 | BIT31);
   
   	/* Config GPP interrupt controller to respond to level trigger */
   	MV_SET_REG_BITS(MV64360_CUNIT_ARBITER_CONTROL_REG, BIT10);

   	/*
    	 * Dismiss and then enable interrupt on GPP interrupt cause for CPU #0
    	 */
   	MV_REG_WRITE(MV64360_GPP_INTERRUPT_CAUSE,
			~(BIT25 | BIT26 | BIT28 | BIT29 | BIT30 | BIT31));
   	MV_SET_REG_BITS(MV64360_GPP_INTERRUPT_MASK0,
			BIT25 | BIT26 | BIT28 | BIT29 | BIT30 | BIT31);

   	/*
    	 * Dismiss and then enable interrupt on CPU #0 high cause register
    	 * BIT27 summarizes GPP interrupts 24-31
    	*/
   	MV_SET_REG_BITS(MV64360_CPU_INTERRUPT0_MASK_HIGH, BIT27);

   	if (ppc_md.progress) 
		ppc_md.progress("chestnut_setup_bridge: exit", 0);
}

/**************************************************************************
 * FUNCTION: chestnut_setup_arch
 *
 * DESCRIPTION: ppc_md machine configuration callback
 *
 ****/
static void __init
chestnut_setup_arch(void)
{
	if (ppc_md.progress)
      		ppc_md.progress("chestnut_setup_arch: enter", 0);

	/* init to some ~sane value until calibrate_delay() runs */
	loops_per_jiffy = 50000000 / HZ;

   	/* if the time base value is greater than bus freq/4 (the TB and
    	* decrementer tick rate) + signed integer rollover value, we
    	* can spend a fair amount of time waiting for the rollover to
    	* happen.  To get around this, initialize the time base register
    	* to a "safe" value.
    	*/
   	set_tb(0, 0);

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0); /* /dev/ram */
	else
#endif
#ifdef CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x0801);	/* /dev/sda1 */
#endif
	
   	/*
    	* Set up the L2CR register.
    	*/
 	_set_L2CR(_get_L2CR() | L2CR_L2E);

	chestnut_setup_bridge();

#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif
#if defined(CONFIG_SERIAL)
	chestnut_early_serial_map();
#endif

	/* Identify the system */
	printk(KERN_INFO "System Identification: IBM 750FX/GX Eval Board\n");
	printk(KERN_INFO "IBM 750FX/GX port (C) 2004 MontaVista Software, Inc. (source@mvista.com)\n");

	if (ppc_md.progress)
      		ppc_md.progress("chestnut_setup_arch: exit", 0);

	return;
}


/**************************************************************************
 * FUNCTION: str_to_hex
 *
 * DESCRIPTION: chestnut_get_mack helper 
 *
 */
static unsigned int 
str_to_hex(char ch)
{
	if ((ch >= '0') && (ch <= '9'))
	    	return(ch - '0');

	else if ((ch >= 'a') && (ch <= 'f'))
	    	return(ch - 'a' + 10);

	else if ((ch >= 'A') && (ch <= 'F'))
	   	 return(ch - 'A' + 10);

	else
		return(0);
}

/**************************************************************************
 * FUNCTION: chestnut_get_mac
 *
 * DESCRIPTION: retrieves the mac address of the interface from the bios
 * 		this was determined by looking at PIBS code from the
 * 		accompanying 750FX Kit CDROM (v. 1.00.0004)
 * 		addr is location of PIBS database - 3rd & 4th entries (256KB
 * 		per entry)
 ***/
void 
chestnut_get_mac(int port_num, char *mac_addr)
{
	int i;
	unsigned long base_addr;
	char addr[12];

	for (i=0; i < 12; i++)
		addr[i]=(0xf);

	if (port_num == 0) 
		base_addr = boot_base+CHESTNUT_PIBS_DATABASE+0x200;
	else
		base_addr = boot_base+CHESTNUT_PIBS_DATABASE+0x300;

	for (i=0; i < 12; i++)
		addr[i] = readb(base_addr+i);

	mac_addr[0] = (str_to_hex(addr[0]) << 4) + 
			str_to_hex(addr[1]);
	mac_addr[1] = (str_to_hex(addr[2]) << 4) + 
			str_to_hex(addr[3]);
	mac_addr[2] = (str_to_hex(addr[4]) << 4) + 
			str_to_hex(addr[5]);
	mac_addr[3] = (str_to_hex(addr[6]) << 4) + 
			str_to_hex(addr[7]);
	mac_addr[4] = (str_to_hex(addr[8]) << 4) + 
			str_to_hex(addr[9]);
	mac_addr[5] = (str_to_hex(addr[10]) << 4) + 
			str_to_hex(addr[11]);
}

/**************************************************************************
 * FUNCTION: chestnut_restart
 * 
 * DESCRIPTION: ppc_md machine reset callback
 *              reset the board via the CPLD command register
 *
 ****/
static void
chestnut_restart(char *cmd)
{
	volatile ulong i = 10000000;

	__cli();

        /*
         * Set CPLD Reg 3 bit 0 to 1 to allow MPP signals on reset to work
         *
         * MPP24 - board reset 
         */
   	writeb(0x1, cpld_base+3);
	
	/* GPP pin tied to MPP earlier */
        MV_SET_REG_BITS(MV64360_GPP_VALUE_SET, BIT24);

   	while (i-- > 0);
   	panic("restart failed\n");
}

static void
chestnut_halt(void)
{
	__cli();
	for (;;);
	/* NOTREACHED */
}

static void
chestnut_power_off(void)
{
	chestnut_halt();
	/* NOTREACHED */
}

/**************************************************************************
 * FUNCTION: chestnut_map_io
 *    
 * DESCRIPTION: configure fixed memory-mapped IO
 *              
 ****/
static void __init
chestnut_map_io(void)
{
#ifdef CONFIG_MV64360_SRAM_CACHEABLE
   	io_block_mapping(CHESTNUT_INTERNAL_SRAM_BASE,
       			CHESTNUT_INTERNAL_SRAM_BASE,
       			CHESTNUT_INTERNAL_SRAM_SIZE,  
       			_PAGE_KERNEL | _PAGE_GUARDED);
#else
#ifdef CONFIG_MV64360_SRAM_CACHE_COHERENT
   	io_block_mapping(CHESTNUT_INTERNAL_SRAM_BASE,
       			CHESTNUT_INTERNAL_SRAM_BASE,
       			CHESTNUT_INTERNAL_SRAM_SIZE,
       			_PAGE_KERNEL | _PAGE_GUARDED | _PAGE_COHERENT);
#else
   	io_block_mapping(CHESTNUT_INTERNAL_SRAM_BASE,
       			CHESTNUT_INTERNAL_SRAM_BASE,
       			CHESTNUT_INTERNAL_SRAM_SIZE,
       			_PAGE_IO);
#endif /* !CONFIG_MV64360_SRAM_CACHE_COHERENT */
#endif /* !CONFIG_MV64360_SRAM_CACHEABLE */

#if defined (CONFIG_SERIAL_TEXT_DEBUG) || defined (CONFIG_KGDB)
	io_block_mapping(CHESTNUT_UART_BASE, CHESTNUT_UART_BASE, 0x100000, _PAGE_IO);
#endif
}

/**************************************************************************
 * FUNCTION: chestnut_set_bat
 *
 * DESCRIPTION: configures a (temporary) bat mapping for early access to 
 *              device I/O
 *              
 ****/
static __inline__ void
chestnut_set_bat(void)
{
        mb();
        mtspr(DBAT3U, 0xf0001ffe);
        mtspr(DBAT3L, 0xf000002a);
        mb();

	return;
}

/**************************************************************************
 * FUNCTION: platform_init
 *
 * DESCRIPTION: main entry point for configuring board-specific machine
 *              callbacks
 *              
 ****/
void __init
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
	parse_bootinfo(find_bootinfo());

#if defined(CONFIG_BLK_DEV_INITRD)
        /*
         * If the init RAM disk has been configured in, and there's a valid
         * starting address for it, set it up.
         */
        if (r4) {
                initrd_start = r4 + KERNELBASE;
                initrd_end = r5 + KERNELBASE;
        }
#endif                          /* CONFIG_BLK_DEV_INITRD */

        /* Copy the kernel command line arguments to a safe place. */

        if (r6) {
                *(char *) (r7 + KERNELBASE) = 0;
                strcpy(cmd_line, (char *) (r6 + KERNELBASE));
        }

	isa_mem_base = 0;

	ppc_md.setup_arch = chestnut_setup_arch;
	ppc_md.show_cpuinfo = chestnut_show_cpuinfo;
	ppc_md.irq_cannonicalize = NULL;
	ppc_md.init_IRQ = mv64360_init_irq;
	ppc_md.get_irq = mv64360_get_irq;
	ppc_md.init = NULL;

	ppc_md.find_end_of_memory = chestnut_find_end_of_memory;
	ppc_md.setup_io_mappings  = chestnut_map_io;

	ppc_md.restart = chestnut_restart;
   	ppc_md.power_off = chestnut_power_off;
   	ppc_md.halt = chestnut_halt;

	ppc_md.time_init = NULL;
	ppc_md.set_rtc_time = NULL;
	ppc_md.get_rtc_time = NULL;
	ppc_md.calibrate_decr = chestnut_calibrate_decr;

	ppc_md.nvram_read_val = NULL;
	ppc_md.nvram_write_val = NULL;

	ppc_md.heartbeat = NULL;

	ppc_md.pcibios_fixup = mv64360_pcibios_fixup;

	chestnut_set_bat();

#if defined(CONFIG_SERIAL) 
#ifdef CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = gen550_progress;
#endif /* CONFIG_SERIAL_TEXT_DEBUG */
	ppc_md.early_serial_map = chestnut_early_serial_map;
#endif

	if (ppc_md.progress)
                ppc_md.progress("chestnut_init(): exit", 0);

        return;
}
