/*
 * arch/ppc/platforms/mv_pmc8245.c
 *
 * Board setup routines for the Marvell MV_PMC8245 Test Platform.
 *
 * Author: Aharon Gadot aharong@radlan.com
 *
 * 2004 (c) Radlan A Marvell company ltd.
 *
 * This file is licensed under
 * the terms of the GNU General Public License version 2.  This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 */

/*
 * This file adds support for the Marvell MV_PMC8245 Test Platform.
 * These board is PMC CPU board with:
 * Motorola PowerPC processor MPC8245LZU266 CPU, 
 * Micron MT48LC8M16A2-7E 128 Mbit SDRAM (64 Mbyte),
 * Two 16 Mbyte Flash devices Intel TE28F256J3C (32 Mbyte total - 16 bit),
 * AMD 29LV040B 512 Bytes boot flash,
 * Two Nation Semiconductors DP83815DUJB ethernet devices (only one connected
 *  to phy - another one for MII interface) - connected via PCI bus,
 * Texas Instruments TL16C550C UART device with 1.8432 Mhz clock,
 * Dallas DS1501WE RTC and NVRAM,
 * Altera EPM3256ATC144-10 CPLD.
 *
 * The boot on the board is U-Boot 1.1.1.  This port depends on this boot
 * to do some basic initialization (e.g., initialize the memory
 * ctlr) and to ensure that the processor is using MAP B (CHRP map).
 *
 * The interrupts are then:
 *	EPIC	Source              Level
 *	0	    PCI INITD# (XINT3)  active low
 *	1	    RTC                 active low
 *	2	    UART                active low
 *	3	    ETHERNET            active low
 *	4	    EXTERNAL INT        active low
 *
 * This is the CPU physical memory map (windows must be at least 1MB and start
 * on a boundary that is a multiple of the window size):
 *
 *   Local Memory       0x00000000 - 0x6FFFFFFF		CS0*    	2G	
 *   SDRAM	            0x00000000 - 0x07FFFFFF		CS0*	    128M    
 *   Flash(16 bit)	    0x7000,0000 -0x707FFFFF	    /FLASH_CS	32M	 - Controlled by RCS2*
 *   PCI region	        0x80000000 - 0xFDFFFFFF	   	            2G	
 *   UART (8  bit)	    0xFF000000 - 0xFF00FFFF	    UART_CS	    64K
 *   CPLD (8  bit)	    0xFF010000 - 0xFF0AFFFF	    CPLD_CS	    192K
 *   RTC  (8  bit)	    0xFF040000 - 0xFF0BFFFF	    RTC_CS	    256Byte
 *   ExBus(8  bit)	    0xFF000000 - 0xFF00FFFF	    EXTD_CS	    192K
 *   Flash/Boot	        0xFF000000 - 0xFFFFFFFF		/FLASH_CS	32M	 - Chip select created from RCS2
 *   BootFlash (8 bit)	0xFFF00000 - 0xFFF7FFFF     /Boot_CS	512k
 *   Reset 	            0xFF010000	
 *   LED	            0xFF010001
 *   Int Mask	        0xFF010002
 *   Int Status	        0xFF010003
 *   Int Cause 	        0xFF010004
 *
 */
 
#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/ide.h>
#include <linux/irq.h>
#include <linux/seq_file.h>
#include <linux/serial.h>
#include <linux/pci_ids.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/time.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/prom.h>
#include <asm/smp.h>
#include <asm/open_pic.h>
#include <asm/todc.h>
#include <asm/bootinfo.h>
#include <asm/mpc10x.h>
#include <asm/pci-bridge.h>
#include <asm/ppcboot.h>

#include "mv_pmc8245_serial.h"
#include "mv_pmc8245.h"

extern void gen550_progress(char *, unsigned short);
extern void gen550_init(int, struct serial_struct *);

unsigned char __res[sizeof (bd_t)];

static void mv_pmc8245_halt(void);

static char *nvBase;
/*
 * Define all of the IRQ senses and polarities.
 */
static u_char mv_pmc8245_openpic_initsenses[] __initdata = {
	(IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE), /* 0: PCI INITD# (XINT3) */
	(IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE), /* 1: RTC                */
	(IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE), /* 2: UART               */
	(IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE), /* 3: ETHERNET           */
	(IRQ_SENSE_LEVEL | IRQ_POLARITY_NEGATIVE), /* 4: EXTERNAL INT       */
};

/*
 * PCI Interrupt routing.
 */
static inline int
mv_pmc8245_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{

    if ((dev->vendor == PCI_VENDOR_ID_NS) &&
        (dev->device == PCI_DEVICE_ID_NS_83815))
    return MV_PMC8245_ETHERNET_IRQ;
    
    return MV_PMC8245_PCI_IRQ;
}
       
static void __init
mv_pmc8245_find_bridges(void)
{
	struct pci_controller *hose;

	hose = pcibios_alloc_controller();

	if (!hose)
		return;

	hose->first_busno = 0;
	hose->last_busno  = 0xff;

	if (mpc10x_bridge_init(hose,
			       MPC10X_MEM_MAP_B,
			       MPC10X_MEM_MAP_B, MPC10X_MAPB_EUMB_BASE) == 0) {

		/* Scan PCI bus */
		hose->last_busno = pciauto_bus_scan(hose, hose->first_busno);

		ppc_md.pcibios_fixup     = NULL;
		ppc_md.pcibios_fixup_bus = NULL;
		ppc_md.pci_swizzle       = common_swizzle;
		ppc_md.pci_map_irq       = mv_pmc8245_map_irq;
	} else {
		if (ppc_md.progress)
			ppc_md.progress("Bridge init failed", 0x100);
		printk("Host bridge init failed\n");
	}

	return;
}

#ifdef CONFIG_SERIAL
unsigned int mv_pmc8245_serial = 0xff000000;

static void __init
mv_pmc8245_early_serial_map(void)
{
	struct serial_struct serial_req;

	/* Setup serial port access */
	memset(&serial_req, 0, sizeof (serial_req));

	serial_req.baud_base        = BASE_BAUD;
	serial_req.line             = 0;
	serial_req.port             = 0;
	serial_req.irq              = MV_PMC8245_SERIAL_0_IRQ;
	serial_req.flags            = STD_COM_FLAGS;
	serial_req.io_type          = SERIAL_IO_MEM;
	serial_req.iomem_reg_shift  = 0;

    serial_req.iomem_base = ioremap(MV_PMC8245_SERIAL_0, 0x20);  

    mv_pmc8245_serial = serial_req.iomem_base;
#if defined(CONFIG_SERIAL_TEXT_DEBUG) || defined(CONFIG_KGDB)
	gen550_init(0, &serial_req);
#endif

#if 0
	if (early_serial_setup(&serial_req) != 0)
		printk("Early serial init of port 0 failed\n");
#endif

}
#endif

static void __init
mv_pmc8245_setup_arch(void)
{
	loops_per_jiffy = 100000000 / HZ;

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	else
#endif
#ifdef	CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x0301);	/* /dev/hda1 IDE disk */
#endif

	/* Lookup PCI host bridges */
	mv_pmc8245_find_bridges();

#ifdef CONFIG_SERIAL
	mv_pmc8245_early_serial_map();
#endif

	nvBase = ioremap(MV_PMC8245_TODC_BASE, MV_PMC8245_TODC_SIZE);
	TODC_INIT(TODC_TYPE_DS1501, 0, 0, nvBase, 8);

#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif

	printk(KERN_INFO "Marvell MV_PMC8245 Test Platform\n");

}
 
static void __init
mv_pmc8245_init2(void)
{
	/* Do board specific initialization.  */

	request_region(0x00, 0x20, "dma1");
	request_region(0x20, 0x20, "pic1");
	request_region(0x40, 0x20, "timer");
	request_region(0x80, 0x10, "dma page reg");
	request_region(0xa0, 0x20, "pic2");
	request_region(0xc0, 0x20, "dma2");

	return;
}

/*
 * Interrupt setup and service. 
 */
static void __init
mv_pmc8245_init_IRQ(void)
{
    int i;

	OpenPIC_InitSenses    = mv_pmc8245_openpic_initsenses;
	OpenPIC_NumInitSenses = sizeof (mv_pmc8245_openpic_initsenses);

	/*
	 * We need to tell openpic_set_sources where things actually are.
	 * mpc10x_common will setup OpenPIC_Addr at ioremap(EUMB phys base +
	 * EPIC offset (0x40000));  The EPIC IRQ Register Address Map -
	 * Interrupt Source Configuration Registers gives these numbers
	 * as offsets starting at 0x50200, we need to adjust occordinly.
	 */

     /* Map 5 external IRQs */
     openpic_set_sources(MV_PMC8245_FIRST_IRQ, 
                         MV_PMC8245_NUM_OF_IRQS, 
			 OpenPIC_Addr + 0x10200);

     /* init openpic wtih first IRQ number - 0 */
     openpic_init(MV_PMC8245_FIRST_IRQ);
}

static u32
mv_pmc8245_irq_cannonicalize(u32 irq)
{
    return irq;
}

static unsigned long __init
mv_pmc8245_find_end_of_memory(void)
{
	return 64 * 1024 * 1024;
}

static void __init
mv_pmc8245_map_io(void)
{
	io_block_mapping(0xfe000000, 0xfe000000, 0x02000000, _PAGE_IO);
}

static void
mv_pmc8245_restart(char *cmd)
{
#if 0
    char *reset;

	__cli();

    /* remap reset to receive virtual adddress */
    reset = (char *)ioremap(CPLD_RESET, 4096);

    *reset = CPLD_RESET_VAL;
#else
  printk (KERN_CRIT "Please cycle power to complete Reset.\n");
#endif

	for (;;) ;		/* Spin until reset happens */
}

static void
mv_pmc8245_power_off(void)
{
	__cli();
	for (;;) ;		/* No way to shut power off with software */
	/* NOTREACHED */
}

static void
mv_pmc8245_halt(void)
{
	mv_pmc8245_power_off();
	/* NOTREACHED */
}

static int
mv_pmc8245_show_cpuinfo(struct seq_file *m)
{
	seq_printf(m, "vendor\t\t: %s\n" ,BOARD_VENDOR);
	seq_printf(m, "machine\t\t: %s\n",BOARD_MACHINE);

	return 0;
}

unsigned char
nvram_read_byte(int addr)
{
  char *cPtr;

 cPtr = nvBase + MV_PMC8245_NVRAM_EXT_RAM_ADRS;
 *cPtr = addr;
 cPtr = nvBase + MV_PMC8245_NVRAM_EXT_RAM_DATA;
 return (*cPtr);
}

void
nvram_write_byte(unsigned char val, int addr)
{
  char *cPtr;

 cPtr = nvBase + MV_PMC8245_NVRAM_EXT_RAM_ADRS;
 *cPtr = addr;
 cPtr = nvBase + MV_PMC8245_NVRAM_EXT_RAM_DATA;
 *cPtr = val;
}

/*
 * Set BAT 3 to map 0xf8000000 to end of physical memory space 1-to-1.
 */
static __inline__ void
mv_pmc8245_set_bat(void)
{
	mb();
	mtspr(DBAT1U, 0xf8000ffe);
	mtspr(DBAT1L, 0xf800002a);
	mb();
}

TODC_ALLOC();

void __init 
platform_init(unsigned long r3, unsigned long r4, unsigned long r5,
	      unsigned long r6, unsigned long r7)
{
	parse_bootinfo(find_bootinfo());

	/* ASSUMPTION:  If both r3 (bd_t pointer) and r6 (cmdline pointer)
	 * are non-zero, then we should use the board info from the bd_t
	 * structure and the cmdline pointed to by r6 instead of the
	 * information from birecs, if any.  Otherwise, use the information
	 * from birecs as discovered by the preceeding call to
	 * parse_bootinfo().  This rule should work with both PPCBoot, which
	 * uses a bd_t board info structure, and the kernel boot wrapper,
	 * which uses birecs.
	 */
	if (r3 && r6) {
		/* copy board info structure */
		memcpy((void *) __res, (void *) (r3 + KERNELBASE),
		       sizeof (bd_t));
		/* copy command line */
		*(char *) (r7 + KERNELBASE) = 0;
		strcpy(cmd_line, (char *) (r6 + KERNELBASE));
	}
#ifdef CONFIG_BLK_DEV_INITRD
	/* take care of initrd if we have one */
	if (r4) {
		initrd_start = r4 + KERNELBASE;
		initrd_end = r5 + KERNELBASE;
	}
#endif				/* CONFIG_BLK_DEV_INITRD */

	/* Map in board regs, etc. */
	mv_pmc8245_set_bat();

	isa_io_base = MPC10X_MAPB_ISA_IO_BASE;
	isa_mem_base = MPC10X_MAPB_ISA_MEM_BASE;
	pci_dram_offset = MPC10X_MAPB_DRAM_OFFSET;
	ISA_DMA_THRESHOLD = 0x00ffffff;
	DMA_MODE_READ = 0x44;
	DMA_MODE_WRITE = 0x48;

	ppc_md.setup_arch = mv_pmc8245_setup_arch;
	ppc_md.show_cpuinfo = mv_pmc8245_show_cpuinfo;
	ppc_md.irq_cannonicalize = mv_pmc8245_irq_cannonicalize;
	ppc_md.init_IRQ = mv_pmc8245_init_IRQ;
	ppc_md.get_irq = openpic_get_irq;
	ppc_md.init = mv_pmc8245_init2;

	ppc_md.restart = mv_pmc8245_restart;
	ppc_md.power_off = mv_pmc8245_power_off;
	ppc_md.halt = mv_pmc8245_halt;

	ppc_md.find_end_of_memory = mv_pmc8245_find_end_of_memory;
	ppc_md.setup_io_mappings = mv_pmc8245_map_io;

	
	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;
	ppc_md.calibrate_decr = todc_calibrate_decr;

	ppc_md.nvram_read_val = todc_direct_read_val;
	ppc_md.nvram_write_val = todc_direct_read_val;

#ifdef CONFIG_SERIAL
#ifdef CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = gen550_progress;
#endif
	ppc_md.early_serial_map = mv_pmc8245_early_serial_map;
#endif

	return;
}
