/*
 * PCI setup routines for the Broadcom BCM5836 family
 *
 * Author: John W. Linville
 *         linville@lvl7.com
 *
 * Copyright 2003 LVL7 Systems, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <common.h>
#include <asm/io.h>
#include "bcm5836.h"

#ifdef CONFIG_PCI

#include <pci.h>

#define PCI_ACCESS_READ  0
#define PCI_ACCESS_WRITE 1

#undef DEBUG

#ifdef DEBUG
#define DBG(x...) printf(x)
#else
#define DBG(x...)
#endif

ulong bcm5836_intmask = 0x00000003;
ulong bcm5836_idmask = 0x0000ffff;

/*
 * Typically there is one core config routine which the pci_ops
 * functions call.
 */
static int
bcm5836_config_access(unsigned char access_type, pci_dev_t bdf,
                      unsigned char where, u32 *data)
{
	u32 addr;

	/*
	 * IMPORTANT
	 * If a pci config cycle fails, it's *very* important
	 * that if the cycle requested is READ, you set *data
	 * to 0xffffffff. The pci_auto code does not check the
	 * return value of the pci_ops functions. It expects that
	 * if a pci config cycle read fails, the data returned
	 * will be 0xffffffff.
	 */

	DBG("%s:%d Configuring Device %d\n", __FUNCTION__, __LINE__, PCI_DEV(bdf));

	if (PCI_BUS(bdf) == 0) {

		if (((1 << PCI_DEV(bdf)) & bcm5836_idmask) == 0) {
    		if (access_type == PCI_ACCESS_READ) {
				*data = 0xffffffff;
			}
	        DBG("%s:%d\n", __FUNCTION__, __LINE__);
			return (-1);
		}

        addr  = (1 << (PCI_DEV(bdf) + 16));

#if 0
		udelay(100);
		writel(((addr & 0xfc000000) | 0x00000002), 0xb8004104);
		udelay(100);
#endif /* bgh */

		addr &= 0x03ffffff;
		addr |= (PCI_FUNC(bdf) << 8);
        addr |= BCM5836_PCI_CFG;
        addr |= (where & 0xfc);

	} else {

		udelay(100);
		writel(0x00000003, 0xb8004104);
		udelay(100);

		addr  = 0x0c000000;
		addr |= bdf;
		addr |= (where & 0xfc);
	}

	addr = KSEG1ADDR(addr);
    DBG("%s:%d register address 0x%x\n", __FUNCTION__, __LINE__, addr);

	if (access_type == PCI_ACCESS_READ) {
#if 0
		if (get_dbe(*data, (u32 *)addr)) {
			*data = 0xffffffff;
			return (-1);
		}
#else
		udelay(100);
		*data = readl(addr);
		udelay(100);
#endif

	} else {
		udelay(100);
		writel(*data, addr);
		udelay(100);
	}

    DBG("%s:%d\n", __FUNCTION__, __LINE__);
		
	return (0);
}


static int
read_config_byte (struct pci_controller *host, pci_dev_t bdf, int where, u8 *val)
{
	u32 data = 0;

	DBG("%s:%d\n", __FUNCTION__, __LINE__);
    
    DBG("cfg read byte: bus %d dev %x fn %x where %x\n", 
         PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where);

    if (bcm5836_config_access(PCI_ACCESS_READ, bdf, where, &data)) {
		*val = 0xff;
		return -1;
	}

	*val = (data >> ((where & 3) << 3)) & 0xff;
    
    DBG("cfg read byte: bus %d dev %x fn %x where %x: val %x\n", 
         PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where, *val);

	return 0;
}


static int
read_config_word (struct pci_controller *host, pci_dev_t bdf, int where, u16 *val)
{
	u32 data = 0;

	DBG("%s:%d\n", __FUNCTION__, __LINE__);
    
    DBG("cfg read word: bus %d dev %x fn %x where %x\n", 
         PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where);

	if (where & 1)
		return -1;

	if (bcm5836_config_access(PCI_ACCESS_READ, bdf, where, &data)) {
		*val = 0xffff;
		return -1;
	}

	*val = (data >> ((where & 3) << 3)) & 0xffff;
    
    DBG("cfg read word: bus %d dev %x fn %x where %x: val %x\n", 
         PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where, *val);

	return 0;
}

static int
read_config_dword (struct pci_controller *host, pci_dev_t bdf, int where, u32 *val)
{
	u32 data = 0;

	DBG("%s:%d\n", __FUNCTION__, __LINE__);
    
    DBG("cfg read dword: bus %d dev %x fn %x where %x\n", 
         PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where);

	if (where & 3)
		return -1;
	
	if (bcm5836_config_access(PCI_ACCESS_READ, bdf, where, &data)) {
		*val = 0xffffffff;
		return -1;
	}

	*val = data;
    
    DBG("cfg read dword: bus %d dev %x fn %x where %x: val %x\n", 
        PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where, *val);

	return 0;
}


static int
write_config_byte (struct pci_controller *host, pci_dev_t bdf, int where, u8 val)
{
	u32 data = 0;
       
	DBG("%s:%d\n", __FUNCTION__, __LINE__);
    
	if (bcm5836_config_access(PCI_ACCESS_READ, bdf, where, &data))
		return -1;

	data = (data & ~(0xff << ((where & 3) << 3))) |
	       (val << ((where & 3) << 3));
    
    DBG("cfg write byte: bus %d dev %x fn %x where %x: val %x\n", 
        PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where, val);

	if (bcm5836_config_access(PCI_ACCESS_WRITE, bdf, where, &data))
		return -1;

	return 0;
}

static int
write_config_word (struct pci_controller *host, pci_dev_t bdf, int where, u16 val)
{
        u32 data = 0;

	DBG("%s:%d\n", __FUNCTION__, __LINE__);
    
	if (where & 1)
		return -1;
       
        if (bcm5836_config_access(PCI_ACCESS_READ, bdf, where, &data))
	       return -1;

	data = (data & ~(0xffff << ((where & 3) << 3))) | 
	       (val << ((where & 3) << 3));
    
    DBG("cfg write word: bus %d dev %x fn %x where %x: val %x\n", 
        PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where, val);

	if (bcm5836_config_access(PCI_ACCESS_WRITE, bdf, where, &data))
	       return -1;


	return 0;
}

static int
write_config_dword(struct pci_controller *host, pci_dev_t bdf, int where, u32 val)
{
	DBG("%s:%d\n", __FUNCTION__, __LINE__);
    
	if (where & 3)
		return -1;

	if (bcm5836_config_access(PCI_ACCESS_WRITE, bdf, where, &val))
	       return -1;
    
    DBG("cfg write dword: bus %d dev %x fn %x where %x: val %x\n", 
        PCI_BUS(bdf), PCI_DEV(bdf), PCI_FUNC(bdf), where, val);

	return 0;
}

void pci_bcm5836_init (struct pci_controller *hose)
{
/*JWL -- need to scan for PCI core address... */
	/* PCI init */

    udelay(100);

	/* Enable PCI tri-states for RST and CLK */
	*((volatile ulong *)0xb8004000) = 0x00000005;
	
	udelay(100);

	/* Enable PCI CLK */
	*((volatile ulong *)0xb8004000) = 0x0000000d;
	
	udelay(100);

	/* Deassert PCI RST */
	*((volatile ulong *)0xb8004000) = 0x0000000f;
	
	udelay(100);

	/* Select internal PCI arbiter */
	*((volatile ulong *)0xb8004010) = 0x0000000d;
	
	udelay(100);

	/* Setup mapping for PCI Memory space */
	*((volatile ulong *)0xb8004100) = 0x08000000;

	udelay(100);

	/* Setup mapping for PCI configuration space */
	*((volatile ulong *)0xb8004104) = 0x00000002;

	udelay(100);

	/* Setup mapping for PCI memory space */
	*((volatile ulong *)0xb8004108) = 0x4000000c;

	udelay(100);

	/* Enable PCI bridge function 0 */
	*((volatile ulong *)0xb8004800) = 0x00000001;

	udelay(100);

	/* Enable PCI bridge BAR 1 */
	*((volatile ulong *)0xb8004810) = 0x0000000f;

	udelay(100);

	/* Initialize bridge command register */

	*((volatile ulong *)0xac010004) = 0xffffffff;

	udelay(100);

	*((volatile ulong *)0xac010004) = 0x00000146;

	udelay(100);

	/* Setup mapping for PCI -> bridge registers */

	/* This is really a dummy, as this feature is not in use... */
	*((volatile ulong *)0xac010010) = 0xb8004000;

	udelay(100);

	*((volatile ulong *)0xac010080) = 0xb8004000;

	udelay(100);

	/* Setup mapping for PCI -> memory */

	*((volatile ulong *)0xac010014) = 0x00000000;

	udelay(100);

	*((volatile ulong *)0xac010084) = 0x00000000;

	udelay(100);

	/* Enable burst & prefetch for PCI -> memory */

	*((volatile ulong *)0xac01008c) = 0x00000003;

	udelay(100);

	/* Set PCI interrupt mask */
	*((volatile ulong *)0xb8004024) = bcm5836_intmask;

	udelay(100);

	hose->first_busno = 0;
	hose->last_busno = 0xff;

	/* System memory space */
	pci_set_region(hose->regions + 0,
		       0x00000000,
		       0x00000000,
		       0x08000000,
		       PCI_REGION_MEM | PCI_REGION_MEMORY);

	/* PCI I/O space */
	pci_set_region(hose->regions + 1,
		       0x00000000,
		       0x08000000,
		       0x04000000,
		       PCI_REGION_IO);

	/* PCI memory space */
	pci_set_region(hose->regions + 2,
		       0x40000000,
		       0x40000000,
		       0x40000000,
		       PCI_REGION_MEM);

	hose->region_count = 3;

	pci_set_ops(hose,
		    read_config_byte,
		    read_config_word,
		    read_config_dword,
		    write_config_byte,
		    write_config_word,
		    write_config_dword);

	pci_register_hose(hose);

	/* Some boards require a HUGE delay... */
	udelay(500000);

	hose->last_busno = pci_hose_scan(hose);

}
#endif
